/* Copyright 1996, Jan D. Wolter and Steven R. Weiss, All Rights Reserved. */

/* Routines to maintain a session database on an SQL server */

#include "backtalk.h"
#include "sysdict.h"

#ifdef ID_SESSION	/* This is all used only with cookie authentication */
char *showopt_session_module= "sess_sql";

#if HAVE_SYS_TIME_H
#include <sys/time.h>   
#else
#include <time.h>
#endif
extern char *crypt (char *key, char *salt);

#include "sess.h"
#include "udb.h"
#include "sql.h"
#include "sqlutil.h"
#include "sqlqry.h"


/* MAKE_SESSION_ID - This generates a 13-character session ID for a new user.
 * The objective here is to construct a string that is (1) likely to be
 * unique among current users, and (2) makes it hard to guess anyone else's
 * session number without actually snooping it.
 *
 * So what we do is construct a "password" based on various data that should
 * be hard to guess for other users and encrypt that with crypt().  We cycle
 * "salt" values sequentially, so the first one issued has salt "AA" and the
 * next "AB" and so on.  This forms the first two characters of the session
 * id and won't repeat for 4,096 users.  This makes having two simultaneous
 * users with the same salt (much less the same session id) pretty unlikely.
 *
 * The "password" is constructed by taking 28 bits from the sum of all
 * login-times of all users since the last time that the session database was
 * emptied.  14 bits are taken from the current second (or millesecond, if
 * available).  The other 14 bits are taken from the process id number, which
 * isn't terribly hard to guess for anyone with an account on the server, but
 * helps blur things a bit more anyway.
 *
 * Since only the encrypted version of the "passwd" is given back to users
 * as a session id, they should not be able to figure out any of the numbers
 * that went into it (which would put them a leg up on guessing what session id
 * the next user might be assigned).  Programs like "crack" won't be very
 * useful in cracking these, since the passwords aren't going to land in any
 * dictionary.  You could crack them with a supercomputer, and then have some
 * data that might help you toward guessing the next session id's but session
 * id's are only good for a few hours anyway, so you'd probably be too late.
 *
 * I'm sure better things could be done, and I'm open to suggestions, but I
 * think this should supply quite adequate security for our needs.
 */

char *make_session_id()
{
    static char *alpha=
	"abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZA./";
    SQLhandle sqlh;
    pid_t pid= getpid();
    int s1, s2;
    char salt[2], raw[8];
    static char *id;
    unsigned long now;

#ifdef HAVE_GETTIMEOFDAY
    struct timeval tv;
    gettimeofday(&tv,NULL);
    now= tv.tv_sec + tv.tv_usec;
#else
    now= time(0L);
#endif

    make_sql_connection();

    sqlh= qry_sess_salt();
    if (sql_nrows(sqlh) == 0)
    	die("make_session_id() could not get salt");
    s1= atol(sql_fetchstr(sqlh,0));
    s2= atol(sql_fetchstr(sqlh,1));
    salt[1]= alpha[s1 % 54];
    salt[0]= alpha[(s1/54) % 54];

    /* Collect 56 bits of gunk to use as our "passwd" */
    raw[0]= s2 % 128;
    raw[1]= (s2 >> 7) % 128;
    raw[2]= (s2 >> 14) % 128;
    raw[3]= (s2 >> 21) % 128;
    raw[4]= now % 128;
    raw[5]= (now >> 7) % 128;
    raw[6]= pid % 128;
    raw[7]= (pid >> 7) % 128;

    /* encrypt it to make a session id */
    id= crypt(raw,salt);

    return id;
}


/* NEW_SESSION - Create new session for the given user.  Pass in a encrypted
 * password and the ip address (in ascii numbers-and-dots form) he connected
 * from.  You should already have checked that it is a valid login.  This will
 * create the session entry in the session database and return the session id
 * in malloced memory.  The caller should set the cookie.
 */

char *new_session(char *user, char *ip)
{
    time_t now= time(0L);
    char *sid;

    /* Generate the new session id and make a copy */
    sid= strdup(make_session_id());

    /* First reap any old sessions */
    qry_sess_reap(now - idict(VAR_EXPIRE_SESSION));

    /* Then create the new one */
    qry_sess_new(sid, user, now, ip);

    return sid;
}


/* SESSION_LOOKUP - Look up the given session id in the session file.  If it
 * is there, and the process id (given here in ascii numbers-and-dots form)
 * matches, then update the last access time in the file and save the user
 * name in the "user" buffer (which should be at least MAX_LOGIN_LEN+1 bytes)
 * and return 0.  Otherwise return 1.
 */

int session_lookup(char *sid, char *ip, char *user)
{
    SQLhandle sqlh;
    char *usr,*ltm,*lip;
    time_t now= time(0L);

    make_sql_connection();

    sqlh= qry_sess_get(sid);
    if (sql_nrows(sqlh) == 0)
    	return 1;

    /* Trap bad data */
    if ((usr= sql_fetchstr(sqlh,0)) == NULL ||
        (ltm= sql_fetchstr(sqlh,1)) == NULL ||
        (lip= sql_fetchstr(sqlh,2)) == NULL)
    	return 1;

    /* Trap expired sessions */
    if (atol(ltm) + idict(VAR_EXPIRE_SESSION) < now)
    	return 1;

#ifdef CHECK_SESSION_IP
    /* Trap IP mismatches */
    if (strcmp(ip,lip))
    	return 1;
#endif

    /* User is OK */
    strncpy(user,usr,MAX_LOGIN_LEN);
    user[MAX_LOGIN_LEN]= '\0';

    qry_sess_refresh(sid, now);

    return 0;
}


/* SESSION_DELETE - Delete the given session ID from the session file, if
 * it exists.  It is not an error if it does not exist.
 */

void session_delete(char *sid)
{
    make_sql_connection();
    qry_sess_kill(sid);
}

#endif /* ID_SESSION */
